"""This is the session controller."""
import asyncio
import logging
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Response
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from starlette.status import (
    HTTP_201_CREATED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_503_SERVICE_UNAVAILABLE,
)

from ...api_models import (
    SessionCreateModel,
    SessionResult,
    SessionResultCollection,
    SessionUpdateModel,
)
from ...database.model import Node, Session, SessionNode, Tenant
from ...database.types import NodeState
from ...nodes_context import contextualize, decontextualize
from ...tasks import deprovision_nodes, fill_pools
from ..auth import req_tenant, req_tenant_optional
from ..database import req_db_async_session

log = logging.getLogger(__name__)

router = APIRouter(prefix="/sessions")


# http get http://localhost:8080/api/v1/sessions
@router.get("", response_model=SessionResultCollection, tags=["sessions"])
async def get_all_sessions(
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Optional[Tenant] = Depends(req_tenant_optional),
):
    """Return all sessions."""
    query = (
        select(Session)
        .options(
            selectinload(Session.tenant),
            selectinload(Session.session_nodes).selectinload(SessionNode.node),
        )
        .filter_by(active=True)
    )
    if tenant and not tenant.is_admin:
        query = query.filter_by(tenant=tenant)
    results = await db_async_session.execute(query)
    return {"action": "get", "sessions": results.scalars().all()}


# http get http://localhost:8080/api/v1/sessions/2
@router.get("/{id}", response_model=SessionResult, tags=["sessions"])
async def get_session(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Return a session with the specified **ID**."""
    session = (
        await db_async_session.execute(
            select(Session)
            .filter_by(id=id)
            .options(
                selectinload(Session.tenant),
                selectinload(Session.session_nodes).selectinload(SessionNode.node),
            )
        )
    ).scalar_one_or_none()
    if not session:
        raise HTTPException(HTTP_404_NOT_FOUND)
    if not tenant.is_admin and session.tenant != tenant:
        raise HTTPException(HTTP_403_FORBIDDEN)
    return {"action": "get", "session": session}


# http --json post http://localhost:8080/api/v1/sessions tenant_id=2 \
#     'nodes_specs:=[{"pool": "virtual-fedora34-x86_64-small", "quantity": 1}]
@router.post("", status_code=HTTP_201_CREATED, response_model=SessionResult, tags=["sessions"])
async def create_session(
    data: SessionCreateModel,
    response: Response,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Create a session with the requested nodes specs."""
    if tenant.is_admin and data.tenant_id is not None:
        tenant = (
            await db_async_session.execute(select(Tenant).filter_by(id=data.tenant_id))
        ).scalar_one_or_none()

        if not tenant:
            raise HTTPException(
                HTTP_422_UNPROCESSABLE_ENTITY, f"can't find tenant with id {data.tenant_id}"
            )
        elif not tenant.active:
            raise HTTPException(
                HTTP_422_UNPROCESSABLE_ENTITY, f"tenant '{tenant.name}' isn't active"
            )
    elif not tenant.is_admin and data.tenant_id is not None and data.tenant_id != tenant.id:
        raise HTTPException(HTTP_403_FORBIDDEN, "can't create session for other tenant")

    session = Session(
        tenant=tenant, data={"nodes_specs": [spec.dict() for spec in data.nodes_specs]}
    )
    db_async_session.add(session)

    nodes_in_transaction = []
    pools_to_fill_up = set()
    for nodes_spec in data.nodes_specs:
        pools_to_fill_up.add(nodes_spec.pool)
        nodes_spec_dict = nodes_spec.dict()
        quantity = nodes_spec_dict.pop("quantity")

        query = (
            select(Node)
            .filter_by(active=True, state=NodeState.ready, **nodes_spec_dict)
            .limit(quantity)
        )

        nodes_to_reserve = (await db_async_session.execute(query)).scalars().all()
        nodes_in_transaction.extend(nodes_to_reserve)

        if len(nodes_to_reserve) < quantity:
            raise HTTPException(HTTP_422_UNPROCESSABLE_ENTITY, f"can't reserve nodes: {nodes_spec}")

        # take the nodes out of circulation and update data
        for node in nodes_to_reserve:
            # record why this node was allocated for this session
            node.data["nodes_spec"] = nodes_spec.dict()
            node.state = NodeState.contextualizing
            session_node = SessionNode(
                session=session, node=node, pool=nodes_spec.pool, data=node.data
            )
            db_async_session.add(session_node)

    await db_async_session.flush()
    nodes_in_transaction = await asyncio.gather(
        *(db_async_session.merge(node, load=False) for node in nodes_in_transaction)
    )
    # Meh. Reload the session instance to ensure all related objects are present in the session.
    session = (
        await db_async_session.execute(
            select(Session)
            .filter_by(id=session.id)
            .options(
                selectinload(Session.tenant),
                selectinload(Session.session_nodes).selectinload(SessionNode.node),
            )
        )
    ).scalar_one()

    contextualized_ipaddrs = await contextualize(
        nodes=[node.ipaddr for node in nodes_in_transaction], ssh_pubkey=tenant.ssh_key
    )

    if None in contextualized_ipaddrs:
        log.error("One or more nodes couldn't be contextualized:")
        nodes_to_decontextualize = []
        for node, ipaddr in zip(nodes_in_transaction, contextualized_ipaddrs):
            if not ipaddr:
                log.error("    id: %s hostname: %s ipaddr: %s", node.id, node.hostname, node.ipaddr)
                node.state = NodeState.failed
                node.data["error"] = "contextualizing node failed"
            else:
                nodes_to_decontextualize.append(node.ipaddr)

        decontextualized_ipaddrs = await decontextualize(nodes=nodes_to_decontextualize)

        if None in decontextualized_ipaddrs:
            log.error("One or more nodes couldn't be decontextualized:")
            for node, ipaddr in zip(nodes_in_transaction, decontextualized_ipaddrs):
                if not ipaddr:
                    log.error(
                        "    id: %s hostname: %s ipaddr: %s", node.id, node.hostname, node.ipaddr
                    )
                    node.state = NodeState.failed
                    node.data["error"] = "decontextualizing node failed"

        await db_async_session.commit()
        response.headers["Retry-After"] = "0"
        raise HTTPException(HTTP_503_SERVICE_UNAVAILABLE, "contextualization of nodes failed")
    else:  # None not in contextualized_ipaddrs
        for node in nodes_in_transaction:
            node.state = NodeState.deployed

    try:
        await db_async_session.commit()
    except IntegrityError as exc:  # pragma: no cover
        raise HTTPException(HTTP_422_UNPROCESSABLE_ENTITY, str(exc))

    # Tell backend worker to fill up pools from which nodes were taken.
    fill_pools.delay(pool_names=list(pools_to_fill_up)).forget()

    return {"action": "post", "session": session}


# http --json put http://localhost:8080/api/v1/sessions/2 active:=false
@router.put("/{id}", response_model=SessionResult, tags=["sessions"])
async def update_session(
    id: int,
    data: SessionUpdateModel,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    session = (
        await db_async_session.execute(
            select(Session)
            .filter_by(id=id)
            .options(
                selectinload(Session.tenant),
                selectinload(Session.session_nodes).selectinload(SessionNode.node),
            )
        )
    ).scalar_one_or_none()

    if not session:
        raise HTTPException(HTTP_404_NOT_FOUND)

    if not tenant.is_admin and session.tenant != tenant:
        raise HTTPException(HTTP_403_FORBIDDEN)

    if not session.active:
        raise HTTPException(HTTP_422_UNPROCESSABLE_ENTITY, f"session {id} is retired")

    if not data.active:
        session.active = data.active
        deprovision_nodes.delay(
            node_ids=[session_node.node_id for session_node in session.session_nodes]
        ).forget()

    await db_async_session.commit()

    return {"action": "put", "session": session}


# http delete http://localhost:8080/api/v1/sessions/2
@router.delete("/{id}", response_model=SessionResult, tags=["sessions"])
async def delete_session(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Delete the session with the specified **ID**."""
    if not tenant.is_admin:
        raise HTTPException(HTTP_403_FORBIDDEN)
    session = (
        await db_async_session.execute(
            select(Session)
            .filter_by(id=id)
            .options(
                selectinload(Session.tenant),
                selectinload(Session.session_nodes).selectinload(SessionNode.node),
            )
        )
    ).scalar_one_or_none()
    if not session:
        raise HTTPException(HTTP_404_NOT_FOUND)
    await db_async_session.delete(session)
    await db_async_session.commit()
    return {"action": "delete", "session": session}



from abc import ABC
from typing import List, Optional

from pydantic import UUID4, BaseModel, SecretStr

from .common import APIResult, CreatableMixin, RetirableMixin

# tenant model


class TenantBase(BaseModel, ABC):
    name: str
    is_admin: Optional[bool]
    ssh_key: SecretStr

    class Config:
        orm_mode = True


class TenantCreateModel(TenantBase):
    pass


class TenantModel(TenantBase, CreatableMixin, RetirableMixin):
    id: int


class TenantCreateResultModel(TenantModel):
    api_key: UUID4


# API results


class TenantResult(APIResult):
    tenant: TenantModel


class TenantCreateResult(TenantResult):
    tenant: TenantCreateResultModel

    class Config:
        json_encoders = {SecretStr: lambda v: v.get_secret_value() if v else None}


class TenantResultCollection(APIResult):
    tenants: List[TenantModel]



import datetime as dt
import enum

from sqlalchemy import Column
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql.expression import FunctionElement
from sqlalchemy.types import DateTime, Enum, SchemaType, TypeDecorator

from ..util import camel_case_to_lower_with_underscores

# adapted from http://techspot.zzzeek.org/2011/01/14/the-enum-recipe
# changes:
# - Python 3.x only
# - uses __init_subclass__() rather than a custom # metaclass
# - improves auto-generated db-side type name
# - is derived from enum.Enum to be usable in pydantic models, therefore doesn't use EnumSymbol


class DeclEnum(enum.Enum):
    """A declarative enumeration type for SQLAlchemy models."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        try:
            del cls._db_type
        except AttributeError:
            pass

        try:
            del cls._literal_values
        except AttributeError:
            pass

    @classmethod
    def db_type(cls):
        if not hasattr(cls, "_db_type"):
            cls._db_type = DeclEnumType(cls)
        else:
            pass  # pragma: no cover
        return cls._db_type

    @classmethod
    def from_string(cls, value):
        try:
            return cls.__members__[value]
        except KeyError:
            raise ValueError(f"Invalid value for {cls.__name__!r}: {value!r}")

    @classmethod
    def values(cls):
        return cls.__members__.keys()


class DeclEnumType(SchemaType, TypeDecorator):
    """A persistable column type tied to a DeclEnum type."""

    cache_ok = True

    def __init__(self, enum):
        self.enum = enum
        self.impl = Enum(*enum.values(), name=self._type_name(enum.__name__))

    @classmethod
    def _type_name(cls, clsname):
        return camel_case_to_lower_with_underscores(clsname) + "_enum"

    def _set_table(self, table, column):
        self.impl._set_table(table, column)

    def copy(self):
        return DeclEnumType(self.enum)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if isinstance(value, str):
            value = self.process_result_value(value, dialect)
        return value.value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self.enum.from_string(value.strip())


# The TZDateTime class is adapted from
# https://docs.sqlalchemy.org/en/14/core/custom_types.html#store-timezone-aware-timestamps-as-timezone-naive-utc
#
# Changes: Is aware of time zones on the database side by default ("explicit better than implicit")


class TZDateTime(TypeDecorator):
    impl = DateTime
    cache_ok = True

    def __init__(self, timezone=True, *args, **kwargs):
        super().__init__(timezone=timezone, *args, **kwargs)

    def process_bind_param(self, value, dialect):
        if value is not None:
            if not value.tzinfo:
                raise TypeError("tzinfo is required")
            value = value.astimezone(dt.timezone.utc).replace(tzinfo=None)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = value.replace(tzinfo=dt.timezone.utc)
        return value


class utcnow(FunctionElement):
    """Current timestamp in UTC for SQL expressions."""

    type = DateTime


@compiles(utcnow, "postgresql")
def _postgresql_utcnow(element, compiler, **kwargs):
    return "(NOW() AT TIME ZONE 'utc')"  # pragma: no cover (unit tests use sqlite)


@compiles(utcnow, "sqlite")
def _sqlite_utcnow(element, compiler, **kwargs):
    return "CURRENT_TIMESTAMP"


# Mixins


class CreatableMixin:
    """An SQLAlchemy mixin to store the time when a thing was created.

    With an asynchronous session this may need to eagerly load the
    default created_at value upon INSERT, e.g. when the attribute is
    accessed on validation of a Pydantic model. This can be achieved by
    setting __mapper_args__ = {"eager_defaults": True}."""

    created_at = Column(TZDateTime, nullable=False, server_default=utcnow())


class RetirableMixin:
    """An SQLAlchemy mixin to manage active state and retirement."""

    retired_at = Column(TZDateTime, nullable=True)

    @hybrid_property
    def active(self) -> bool:
        return self.retired_at is None

    @active.setter
    def active(self, value: bool):
        if not value:
            # only set retired_at if previously unset
            if not self.retired_at:
                self.retired_at = dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)
        else:
            self.retired_at = None

    @active.expression
    def active(cls):
        return cls.retired_at == None  # noqa: E711



import logging
import sys

from fastapi import FastAPI

from .. import database, tasks
from ..exceptions import DuffyConfigurationError
from ..version import __version__
from .controllers import node, session, tenant

log = logging.getLogger(__name__)

description = """
Duffy is the middle layer running [`ci.centos.org`](https://ci.centos.org). It provisions, tears
down and rebuilds physical and virtual machines which are used to run tests in the CentOS CI
Cluster.
"""

tags_metadata = [
    {"name": "nodes", "description": "Operations with physical and virtual nodes"},
    {"name": "sessions", "description": "Operations with sessions"},
    {"name": "tenants", "description": "Operations with tenants"},
]

app = FastAPI(
    title="Duffy",
    description=description,
    version=__version__,
    contact={"name": "CentOS CI", "email": "ci-sysadmin@centos.org"},
    openapi_tags=tags_metadata,
)


# API v1

PREFIX = "/api/v1"

app.include_router(node.router, prefix=PREFIX)
app.include_router(tenant.router, prefix=PREFIX)
app.include_router(session.router, prefix=PREFIX)


# DB model initialization


@app.on_event("startup")
async def init_model():
    try:
        database.init_sync_model()
        await database.init_async_model()
    except DuffyConfigurationError as exc:
        log.error("Configuration key missing or wrong: %s", exc.args[0])
        sys.exit(1)


# Celery tasks initialization


@app.on_event("startup")
def init_tasks():
    tasks.init_tasks()



from enum import Enum, auto
from typing import Any, Dict, List, Optional

import ansible_runner
from celery.utils.log import get_task_logger

from ...database.model import Node
from .main import Mechanism, MechanismFailure

log = get_task_logger(__name__)


class PlaybookType(Enum):
    provision = auto()
    deprovision = auto()


class AnsibleMechanism(Mechanism, mech_type="ansible"):
    def run_playbook(
        self,
        playbook_type: PlaybookType,
        failure_msg: str,
        extra_vars: Optional[Dict[str, Any]] = None,
        overrides: Optional[Dict[str, Any]] = None,
    ):
        log.debug(
            "AnsibleMechanism.run_playbook(%r, %r)\n\t%r\n\t%r",
            self,
            playbook_type,
            extra_vars,
            overrides,
        )
        subconf = self[playbook_type.name]

        run_extra_vars = self.get("extra-vars", {})
        if "extra-vars" in subconf:
            run_extra_vars = {**run_extra_vars, **subconf["extra-vars"]}
        if extra_vars:
            run_extra_vars = {**run_extra_vars, **extra_vars}

        run_extra_vars = self.nodepool.render_templates_in_obj(run_extra_vars, overrides=overrides)

        log.debug(
            "ansible_runner.run(project_dir=%r, playbook=%r, json_mode=True, extravars=%r",
            self["topdir"],
            self[playbook_type.name]["playbook"],
            run_extra_vars,
        )
        run = ansible_runner.run(
            project_dir=self["topdir"],
            playbook=self[playbook_type.name]["playbook"],
            json_mode=True,
            extravars=run_extra_vars,
        )

        if run.status != "successful":
            raise MechanismFailure(failure_msg)

        for event in reversed(list(run.events)):
            event_type = event["event"]
            event_data = event["event_data"]
            event_res = event_data.get("res")

            if (
                event_type == "runner_on_ok"
                and event_data.get("task_action") == "set_fact"
                and "duffy_out" in event_res.get("ansible_facts", {})
            ):
                return event_res["ansible_facts"]["duffy_out"]
        else:
            raise MechanismFailure(failure_msg)

    def provision(self, nodes: List[Node]) -> Dict[str, Any]:
        playbook_input = {
            "duffy_in": {
                "nodes": [
                    {"id": node.id, "hostname": node.hostname, "ipaddr": node.ipaddr}
                    for node in nodes
                ],
            },
        }
        return self.run_playbook(
            PlaybookType.provision,
            "Provisioning failed",
            extra_vars=playbook_input,
            overrides=playbook_input,
        )

    def deprovision(self, nodes: List[Node]) -> Dict[str, Any]:
        playbook_input = {
            "duffy_in": {
                "nodes": [
                    {
                        "id": node.id,
                        "hostname": node.hostname,
                        "ipaddr": node.ipaddr,
                        "data": node.data,
                    }
                    for node in nodes
                ],
            },
        }
        return self.run_playbook(
            PlaybookType.deprovision,
            "Deprovisioning failed",
            extra_vars=playbook_input,
            overrides=playbook_input,
        )



from .common import APIResult, APIResultAction  # noqa: F401
from .node import NodeCreateModel, NodeModel, NodeResult, NodeResultCollection  # noqa: F401
from .session import (  # noqa: F401
    SessionCreateModel,
    SessionModel,
    SessionNodeModel,
    SessionResult,
    SessionResultCollection,
    SessionUpdateModel,
)
from .tenant import (  # noqa: F401
    TenantCreateModel,
    TenantCreateResult,
    TenantCreateResultModel,
    TenantModel,
    TenantResult,
    TenantResultCollection,
)



from abc import ABC
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, conint

from ..database.types import NodeState
from .common import APIResult, CreatableMixin, RetirableMixin
from .node import NodeBase
from .tenant import TenantModel

# nodes spec model


class NodesSpec(BaseModel):
    quantity: conint(ge=1)
    pool: str


# nodes in sessions model


class SessionNodeModel(NodeBase):
    id: int
    state: Optional[NodeState]


# session model


class SessionBase(BaseModel, ABC):
    class Config:
        orm_mode = True


class SessionCreateModel(SessionBase):
    tenant_id: Optional[int]
    nodes_specs: List[NodesSpec]


class SessionUpdateModel(SessionBase):
    active: bool


class SessionModel(SessionBase, CreatableMixin, RetirableMixin):
    id: int
    tenant: TenantModel
    data: Dict[str, Any]
    nodes: List[SessionNodeModel]


# API results


class SessionResult(APIResult):
    session: SessionModel


class SessionResultCollection(APIResult):
    sessions: List[SessionModel]



from abc import ABC
from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class CreatableMixin(BaseModel):
    created_at: datetime


class RetirableMixin(BaseModel):
    retired_at: Optional[datetime]
    active: bool


class APIResultAction(str, Enum):
    delete = "delete"
    get = "get"
    post = "post"
    put = "put"


class APIResult(BaseModel, ABC):
    action: Optional[APIResultAction]



from typing import Tuple

from celery import Celery

from .. import database
from ..configuration import config
from .base import celery, init_tasks
from .node_pools import NodePool
from .provision import fill_pools

DEFAULT_PERIODIC_INTERVAL = 5 * 60


@celery.on_after_finalize.connect
def setup_periodic_tasks(sender: Celery, **kwargs):
    periodic_config = config.get("tasks", {}).get("periodic", {})

    sender.add_periodic_task(
        periodic_config.get("fill-pools", {}).get("interval", DEFAULT_PERIODIC_INTERVAL),
        fill_pools.signature(),
    )


@celery.on_after_finalize.connect
def run_init_tasks(sender: Celery, **kwargs):
    fill_pools.delay().forget()


def start_worker(worker_args: Tuple[str]):
    database.init_sync_model()
    init_tasks()
    NodePool.process_configuration()

    celery.worker_main(("worker",) + worker_args)



from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from starlette.status import (
    HTTP_201_CREATED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
)

from ...api_models import NodeCreateModel, NodeResult, NodeResultCollection
from ...database.model import Node, Tenant
from ..auth import req_tenant
from ..database import req_db_async_session

router = APIRouter(prefix="/nodes")


@router.get("", response_model=NodeResultCollection, tags=["nodes"])
async def get_all_nodes(
    db_async_session: AsyncSession = Depends(req_db_async_session),
):
    """Return all nodes."""
    query = select(Node).options(selectinload("*")).filter_by(active=True)
    results = await db_async_session.execute(query)

    return {"action": "get", "nodes": results.scalars().all()}


@router.get("/{id}", response_model=NodeResult, tags=["nodes"])
async def get_node(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Return the node with the specified **ID**."""
    query = select(Node).filter_by(id=id).options(selectinload("*"))
    result = await db_async_session.execute(query)
    node = result.scalar_one_or_none()

    if not node:
        raise HTTPException(HTTP_404_NOT_FOUND)

    return {"action": "get", "node": node}


@router.post("", status_code=HTTP_201_CREATED, response_model=NodeResult, tags=["nodes"])
async def create_node(
    data: NodeCreateModel,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Create a node with the requested properties."""
    if not tenant.is_admin:
        raise HTTPException(HTTP_403_FORBIDDEN)

    args = {
        "hostname": data.hostname,
        "ipaddr": str(data.ipaddr),
        "comment": data.comment,
        "pool": data.pool,
        "reusable": data.reusable,
        "data": data.data,
    }

    node = Node(**args)
    db_async_session.add(node)

    try:
        await db_async_session.commit()
    except IntegrityError as exc:  # pragma: no cover
        raise HTTPException(HTTP_409_CONFLICT, str(exc))

    await db_async_session.refresh(node)

    return {"action": "post", "node": node}


@router.delete("/{id}", response_model=NodeResult, tags=["nodes"])
async def delete_node(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Delete the node with the specified **ID**."""
    if not tenant.is_admin:
        raise HTTPException(HTTP_403_FORBIDDEN)

    node = (
        await db_async_session.execute(select(Node).filter_by(id=id).options(selectinload("*")))
    ).scalar_one_or_none()

    if not node:
        raise HTTPException(HTTP_404_NOT_FOUND)

    await db_async_session.delete(node)
    await db_async_session.commit()

    return {"action": "delete", "node": node}



from typing import Optional

import httpx
from fastapi import Depends, FastAPI
from fastapi.responses import JSONResponse
from starlette.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_401_UNAUTHORIZED,
    HTTP_403_FORBIDDEN,
)

from ..configuration import config
from ..version import __version__
from .api_models import Credentials
from .auth import req_credentials, req_credentials_optional

description = """
Duffy is the middle layer running [`ci.centos.org`](https://ci.centos.org). It provisions, tears
down and rebuilds physical and virtual machines are used to run tests in the CentOS CI Cluster.

This metaclient exposes older endpoints for legacy support and connects them to the path operations
introduced by the newer version of the Duffy endpoint, until the support for the older endpoints is
deprecated.
"""

app = FastAPI(
    title="Duffy Metaclient for Legacy Support",
    description=description,
    version=__version__,
    contact={"name": "CentOS CI", "email": "ci-sysadmin@centos.org"},
)


@app.get("/Node/get")
async def get_node(
    ver: str = "7",
    arch: str = "x86_64",
    count: int = 1,
    flavor: str = None,
    cred: Credentials = Depends(req_credentials),
):
    ver = "".join(i.replace("-", "") for i in ver)
    if arch in ["aarch64", "ppc64", "ppc64le"]:
        if not flavor:
            flavor = "medium"
        nodes_specs = [{"quantity": count, "pool": f"virtual-centos{ver}-{arch}-{flavor}"}]
    else:
        nodes_specs = [{"quantity": count, "pool": f"physical-centos{ver}-{arch}"}]

    async with httpx.AsyncClient() as client:
        dest = config["metaclient"]["dest"].rstrip("/")
        response = await client.post(
            f"{dest}/api/v1/sessions",
            json={"nodes_specs": nodes_specs},
            auth=(cred.username, cred.password),
        )

    if response.status_code == HTTP_201_CREATED:
        session = response.json()["session"]
        legacy_result = {
            "ssid": session["id"],
            "hosts": [node["hostname"] for node in session["nodes"]],
        }
        return legacy_result
    else:
        return JSONResponse(status_code=HTTP_200_OK, content="Failed to allocate nodes")


@app.get("/Node/done")
async def return_node_on_completion(ssid: str = None, cred: Credentials = Depends(req_credentials)):
    if ssid:
        async with httpx.AsyncClient() as client:
            dest = config["metaclient"]["dest"].rstrip("/")
            response = await client.put(
                f"{dest}/api/v1/sessions/{ssid}",
                json={"active": False},
                auth=(cred.username, cred.password),
            )
        if response.status_code == HTTP_200_OK:
            return JSONResponse(status_code=HTTP_200_OK, content="Done")
        elif (
            response.status_code == HTTP_401_UNAUTHORIZED
            or response.status_code == HTTP_403_FORBIDDEN
        ):
            return JSONResponse(
                status_code=HTTP_403_FORBIDDEN, content={"msg": "Invalid duffy key"}
            )
        else:
            return JSONResponse(
                status_code=HTTP_200_OK, content="Failed to return nodes on completion"
            )
    else:
        return JSONResponse(status_code=HTTP_200_OK, content="Some parameters are absent")


@app.get("/Node/fail")
async def return_node_on_failure(ssid: str = None, cred: Credentials = Depends(req_credentials)):
    if ssid:
        return JSONResponse(status_code=HTTP_200_OK, content="Not implemented yet")
    else:
        return JSONResponse(status_code=HTTP_200_OK, content="Some parameters are absent")


@app.get("/Inventory")
async def get_nodes(cred: Optional[Credentials] = Depends(req_credentials_optional)):
    if cred:
        auth = cred.username, cred.password
        async with httpx.AsyncClient() as client:
            dest = config["metaclient"]["dest"].rstrip("/")
            response = await client.get(f"{dest}/api/v1/sessions", auth=auth)
        if response.status_code == HTTP_200_OK:
            retnlist = []
            sesslist = response.json()["sessions"]
            for jndx in sesslist:
                nodelist = jndx["nodes"]
                for indx in nodelist:
                    retnlist.append([indx.get("hostname"), jndx.get("id")])
            return JSONResponse(status_code=HTTP_200_OK, content=retnlist)
        elif (
            response.status_code == HTTP_401_UNAUTHORIZED
            or response.status_code == HTTP_403_FORBIDDEN
        ):
            return JSONResponse(
                status_code=HTTP_403_FORBIDDEN, content={"msg": "Invalid duffy key"}
            )
        else:
            # Yes, the legacy API returns 200 here.
            return JSONResponse(
                status_code=HTTP_200_OK, content="Failed to retrieve inventory of nodes"
            )
    else:
        auth = None
        async with httpx.AsyncClient() as client:
            dest = config["metaclient"]["dest"].rstrip("/")
            response = await client.get(f"{dest}/api/v1/sessions", auth=auth)
        if response.status_code == HTTP_200_OK:
            retnlist = []
            sesslist = response.json()["sessions"]
            for jndx in sesslist:
                nodelist = jndx["nodes"]
                for indx in nodelist:
                    retnlist.append(
                        [
                            indx.get("id"),  # sch.data['id']
                            indx.get("hostname"),  # sch.data['hostname']
                            indx.get("ipaddr"),  # sch.data['ip']
                            indx.get("type"),  # sch.data['chassis']
                            0,  # sch.data['used_count']
                            indx.get("state"),  # sch.data['state']
                            indx.get("comment"),  # sch.data['comment']
                            indx.get("distro_type"),  # sch.data['distro']
                            indx.get("distro_version"),  # sch.data['rel']
                            indx.get("distro_version"),  # sch.data['ver']
                            indx.get("arch"),  # sch.data['arch']
                            indx.get("pool"),  # sch.data['pool']
                            indx.get("console_port"),  # sch.data['console_port']
                            indx.get("flavour"),  # sch.data['flavor']
                        ]
                    )
            return JSONResponse(status_code=HTTP_200_OK, content=retnlist)
        else:
            # Yes, the legacy API returns 200 here.
            return JSONResponse(
                status_code=HTTP_200_OK, content="Failed to retrieve inventory of nodes"
            )



def camel_case_to_lower_with_underscores(camelcased: str) -> str:
    """Convert CamelCased names to lower_case_with_underscores."""
    chunk_positions = []
    prev_split_pos = 0

    for pos, (prev_is_upper, char_is_upper, next_is_upper) in enumerate(
        zip(
            (x.isupper() for x in camelcased[:-2]),
            (x.isupper() for x in camelcased[1:-1]),
            (x.isupper() for x in camelcased[2:]),
        ),
        start=1,
    ):
        if char_is_upper and (not prev_is_upper or not next_is_upper):
            chunk_positions.append((prev_split_pos, pos))
            prev_split_pos = pos

    chunk_positions.append((prev_split_pos, len(camelcased)))

    lowercased = camelcased.lower()

    return "_".join(
        lowercased[prev_split_pos:split_pos] for prev_split_pos, split_pos in chunk_positions
    )


def merge_dicts(*src_dicts):
    """Create a deep merge of several dictionaries.

    The structure of the dictionaries must be compatible, i.e. sub-dicts
    may not be differently typed between the source dictionaries."""
    if not src_dicts:
        raise ValueError("Can't merge nothing")

    if not all(isinstance(src_dict, dict) for src_dict in src_dicts):
        raise TypeError("All objects to be merged have to be dictionaries")

    res_dict = {}

    for src_dict in src_dicts:
        for key, src_value in src_dict.items():
            if isinstance(src_value, dict):
                if key not in res_dict:
                    res_dict[key] = src_value.copy()
                else:
                    res_dict[key] = merge_dicts(res_dict[key], src_value)
            elif key in res_dict and isinstance(res_dict[key], dict):
                raise TypeError("All objects to be merged have to be dictionaries")
            else:
                res_dict[key] = src_value

    return res_dict



import copy
import logging
import sys
from typing import Tuple

import click
import uvicorn
import yaml

from . import database, shell
from .configuration import config, read_configuration
from .database.migrations.main import alembic_migration
from .database.setup import setup_db_schema, setup_db_test_data
from .exceptions import DuffyConfigurationError
from .tasks import start_worker
from .version import __version__

DEFAULT_CONFIG_FILE = "/etc/duffy.yaml"

log = logging.getLogger(__name__)


# Global setup and CLI options


def init_config(ctx, param, filename):
    try:
        read_configuration(filename)
    except FileNotFoundError:
        if filename is not DEFAULT_CONFIG_FILE:
            raise


@click.group(name="duffy")
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False),
    default=DEFAULT_CONFIG_FILE,
    callback=init_config,
    is_eager=True,
    expose_value=False,
    help="Read option defaults from the specified YAML file.",
    show_default=True,
)
@click.version_option(version=__version__, prog_name="Duffy")
def cli():
    pass


# Check & dump configuration


@cli.group(name="config")
def config_subcmd():
    """Check and dump configuration."""


@config_subcmd.command(name="check")
def config_check():
    """Validate configuration structure.

    This checks if configuration subkeys conform to the expected format.
    However, it doesn't check if the sub-keys necessary to run a certain
    subcommand exist.
    """
    if not config:
        click.echo("Configuration is empty.")
    else:
        click.echo(f"OK.\nValidated configuration subkeys: {', '.join(config)}")


@config_subcmd.command("dump")
def config_dump():
    """Dump merged configuration."""
    yaml.safe_dump(config, sys.stdout)


# Set up the database tables


@cli.command()
@click.option(
    "--test-data/--no-test-data", default=False, help="Initialized database with test data."
)
def setup_db(test_data):
    """Create tables from the database model."""
    try:
        setup_db_schema()
        if test_data:
            database.init_model()
            setup_db_test_data()
    except DuffyConfigurationError as exc:
        log.error("Configuration key missing or wrong: %s", exc.args[0])
        sys.exit(1)


# Handle database migrations


@cli.group()
def migration():
    """Handle database migrations."""
    pass


@migration.command("create")
@click.option(
    "--autogenerate/--no-autogenerate",
    default=False,
    help="Autogenerate migration script skeleton (needs to be reviewed/edited).",
)
@click.argument("comment", nargs=-1, required=True)
def migration_create(autogenerate, comment):
    """Create a new migration."""
    alembic_migration.create(comment=" ".join(comment), autogenerate=autogenerate)


@migration.command("db-version")
def migration_db_version():
    alembic_migration.db_version()


@migration.command("upgrade")
@click.argument("version", default="head")
def migration_upgrade(version):
    alembic_migration.upgrade(version)


@migration.command("downgrade")
@click.argument("version", default="-1")
def migration_downgrade(version):
    alembic_migration.downgrade(version)


# Interactive shell


@cli.command(name="shell")
@click.option(
    "-t",
    "--shell-type",
    type=click.Choice(shell.get_available_shells(), case_sensitive=False),
    help="Type of interactive shell to use.",
    default=None,
)
def run_shell(shell_type: str):
    """Run an interactive shell."""
    try:
        database.init_model()
    except DuffyConfigurationError as exc:
        log.error("Configuration key missing or wrong: %s", exc.args[0])
        sys.exit(1)

    shell.embed_shell(shell_type=shell_type)


# Run the backend task worker


@cli.command(context_settings={"ignore_unknown_options": True, "help_option_names": []})
@click.argument("worker_args", nargs=-1, type=click.UNPROCESSED)
def worker(worker_args: Tuple[str]):
    """Start a Celery worker to process backend tasks."""
    start_worker(worker_args=worker_args)


# Run the web app


@cli.command()
@click.option(
    "--reload/--no-reload", default=False, help="Automatically reload if the code is changed."
)
@click.option("-H", "--host", default=None, help="Set the host address to listen on.")
@click.option(
    "-p",
    "--port",
    type=click.IntRange(1, 65535, clamp=True),
    default=None,
    help="Set the port value.",
)
@click.option(
    "-l",
    "--loglevel",
    "loglevel",
    type=click.Choice(list(uvicorn.config.LOG_LEVELS.keys()), case_sensitive=False),
    help="Set the log level.",
    default=None,
)
def serve(reload, host, port, loglevel):
    """Run the Duffy web application server.

    Duffy is the middle layer running ci.centos.org that manages the
    provisioning, maintenance, teardown and rebuild of the Nodes
    (physical hardware and virtual machines) that are used to run
    the tests in the CI Cluster.
    """
    config_app = config.get("app", {})
    if host is None:
        host = config_app.get("host", "127.0.0.1")
    if port is None:
        port = config_app.get("port", 8080)
    if loglevel is None:
        loglevel = config_app.get("loglevel", "info")

    numeric_loglevel = uvicorn.config.LOG_LEVELS[loglevel.lower()]
    logging.basicConfig(level=numeric_loglevel)

    # Report for duty
    print(" * Starting Duffy...")
    print(f" * Host address : {host}")
    print(f" * Port number  : {port}")
    print(f" * Log level    : {loglevel}")
    print(f" * Serving API docs on http://{host}:{port}/docs")

    uvicorn_log_config = copy.deepcopy(config_app.get("logging", uvicorn.config.LOGGING_CONFIG))
    if uvicorn_log_config.get("loggers", {}).get("duffy"):
        uvicorn_log_config["loggers"]["duffy"]["level"] = numeric_loglevel

    # Start the show
    uvicorn.run(
        "duffy.app.main:app",
        host=host,
        port=port,
        log_level=numeric_loglevel,
        reload=reload,
        log_config=uvicorn_log_config,
    )


# Run the web app - Duffy Metaclient for Legacy Support


@cli.command()
@click.option(
    "--reload/--no-reload", default=False, help="Automatically reload if the code is changed."
)
@click.option("-H", "--host", default=None, help="Set the host address to listen on.")
@click.option(
    "-p",
    "--port",
    type=click.IntRange(1, 65535, clamp=True),
    default=None,
    help="Set the port value.",
)
@click.option(
    "-D",
    "--dest",
    default=None,
    help="Set the destination address of Duffy deployment.",
)
@click.option(
    "-l",
    "--loglevel",
    "loglevel",
    type=click.Choice(list(uvicorn.config.LOG_LEVELS.keys()), case_sensitive=False),
    help="Set the log level.",
    default=None,
)
@click.pass_context
def serve_legacy(ctx, reload, host, port, dest, loglevel):
    """Serve the Duffy Metaclient for Legacy Support app.

    Duffy is the middle layer running ci.centos.org that manages the
    provisioning, maintenance and teardown / rebuild of the Nodes
    (physical hardware for now, VMs coming soon) that are used to run
    the tests in the CI Cluster.

    This metaclient exposes older endpoints for legacy support and
    connects them to the path operations introduced by the newer version
    of the Duffy endpoint, until the support for the older endpoints is
    deprecated.
    """
    config_metaclient = config.get("metaclient", {})
    if host is None:
        host = config_metaclient.get("host", "127.0.0.1")
    if port is None:
        port = config_metaclient.get("port", 9090)
    if dest is None:
        dest = config_metaclient.get("dest", "http://127.0.0.1:8080")
    if loglevel is None:
        loglevel = config_metaclient.get("loglevel", "info")

    numeric_loglevel = uvicorn.config.LOG_LEVELS[loglevel.lower()]
    logging.basicConfig(level=numeric_loglevel)

    # Report for duty
    print(" * Starting Duffy Metaclient for Legacy Support...")
    print(f" * Host address        : {host}")
    print(f" * Port number         : {port}")
    print(f" * Destination address : {dest}")
    print(f" * Log level           : {loglevel}")
    print(f" * Serving API docs on http://{host}:{port}/docs")

    uvicorn_log_config = copy.deepcopy(
        config_metaclient.get("logging", uvicorn.config.LOGGING_CONFIG)
    )
    if uvicorn_log_config.get("loggers", {}).get("duffy"):
        uvicorn_log_config["loggers"]["duffy"]["level"] = numeric_loglevel

    # Start the show
    uvicorn.run(
        "duffy.legacy.main:app",
        host=host,
        port=port,
        log_level=numeric_loglevel,
        reload=reload,
        log_config=uvicorn_log_config,
    )



from abc import ABC
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, IPvAnyAddress

from ..database.types import NodeState
from .common import APIResult, CreatableMixin, RetirableMixin

# abstract node


class NodeBase(BaseModel, ABC):
    hostname: str
    ipaddr: IPvAnyAddress
    comment: Optional[str]

    pool: Optional[str]
    reusable: bool
    data: Dict[str, Any]

    class Config:
        orm_mode = True


class NodeCreateModel(NodeBase):
    reusable: bool = False
    data: Dict[str, Any] = {}


class NodeModel(NodeBase, CreatableMixin, RetirableMixin):
    id: int
    state: NodeState


# API results


class NodeResult(APIResult):
    node: NodeModel


class NodeResultCollection(APIResult):
    nodes: List[NodeModel]



from typing import List, Optional

from celery.utils.log import get_task_logger
from sqlalchemy import func, select

from ..database import sync_session_maker
from ..database.model import Node
from ..database.types import NodeState
from .base import celery
from .mechanisms import MechanismFailure
from .node_pools import ConcreteNodePool, NodePool

log = get_task_logger(__name__)


@celery.task
def fill_single_pool(pool_name: str):
    try:
        pool = NodePool.known_pools[pool_name]
    except KeyError as exc:
        raise RuntimeError(f"[{pool_name}] Unknown pool, bailing out") from exc

    log.debug("[%s] Filling up pool ...", pool.name)

    # Determine how many nodes need to be provisioned and commit them to the database so these nodes
    # are considered in future calculations.

    wanted_fill_level = pool["fill-level"]

    with sync_session_maker() as db_sync_session, db_sync_session.begin():
        log.debug("[%s] Determining number of available nodes ...", pool.name)
        current_fill_level = db_sync_session.execute(
            select(func.count()).select_from(
                select(Node)
                .filter_by(active=True, pool=pool.name, state=NodeState.ready)
                .subquery()
            )
        ).scalar_one()

        quantity = wanted_fill_level - current_fill_level
        log.debug(
            "[%s] ... %d, we want %d, i.e. need %d",
            pool.name,
            current_fill_level,
            wanted_fill_level,
            quantity,
        )

        if quantity <= 0:
            log.debug("[%s] Pool is filled to or above spec.", pool.name)
            return

        reuse_nodes = pool.get("reuse-nodes")
        if reuse_nodes:
            log.debug("[%s] Searching for %d reusable nodes in database", pool.name, quantity)

            usable_nodes_query = select(Node).filter_by(
                active=True, pool=None, reusable=True, state=NodeState.unused
            )
            spec_valid = True
            for key, value in reuse_nodes.items():
                json_item = Node.data[key]
                if isinstance(value, str):
                    cast_item = json_item.as_string()
                    value = pool.render_template(value)
                elif isinstance(value, int):
                    cast_item = json_item.as_integer()
                else:
                    log.error(
                        "[%s] Can't build query for reuse-nodes: %r -> %r", pool.name, key, value
                    )
                    spec_valid = False
                    continue
                usable_nodes_query = usable_nodes_query.filter(cast_item == value)

            if not spec_valid:
                raise RuntimeError(f"[{pool.name}] Skipping filling up")

            # This queries up to `quantity` usable nodes, or fewer.
            usable_nodes_query = usable_nodes_query.limit(quantity)
            log.debug("Usable nodes query: %s", usable_nodes_query)
            nodes = db_sync_session.execute(usable_nodes_query).scalars().all()
            log.info("Found %d suitable unused nodes.", len(nodes))
        else:
            log.debug("[%s] Allocating %d new node objects in database", pool.name, quantity)
            nodes = [Node() for i in range(quantity)]
            db_sync_session.add_all(nodes)

        for node in nodes:
            node.state = NodeState.provisioning
            node.pool = pool.name

    # In a second (longer running) transaction, provision the nodes.

    with sync_session_maker() as db_sync_session, db_sync_session.begin():
        # Make the nodes available in the new transaction/session.
        nodes = [db_sync_session.merge(node, load=False) for node in nodes]

        # Here, `nodes` is a list of node objects which are in state "provisioning" and assigned
        # to the pool. It can be shorter than `quantity`, e.g. if there aren't enough reusable
        # unused nodes.
        log.info("[%s] Attempting to provision %d nodes ...", pool.name, quantity)
        try:
            prov_result = pool.provision(nodes)
        except MechanismFailure:
            log.error("[%s] Provisioning failed.", pool.name)
            if not reuse_nodes:
                for node in nodes:
                    db_sync_session.delete(node)
            else:
                for node in nodes:
                    node.state = NodeState.unused
                    node.pool = None
            return
        log.info("[%s] Backend provisioning finished.", pool.name)
        log.debug("[%s] Result: %s", pool.name, prov_result)

        if reuse_nodes:
            # As existing nodes are used, hostname and ipaddr (if any) in the results will be
            # ignored, only the size of the result set will be considered.
            valid_node_results = prov_result["nodes"]
        else:
            # The playbook needs to provide hostname and ipaddr for provisioned nodes in the
            # result, otherwise they can't be used.
            log.debug("[%s] Validating backend provisioning node results.", pool.name)
            valid_node_results = []
            invalid_node_results = []
            for node_res in prov_result["nodes"]:
                hostname = node_res.get("hostname")
                ipaddr = node_res.get("ipaddr")

                if not (hostname and ipaddr):
                    invalid_node_results.append(node_res)
                    continue

                valid_node_results.append(node_res)

            log.debug("[%s] valid results: %s", pool.name, valid_node_results)
            log.debug("[%s] invalid results: %s", pool.name, invalid_node_results)

            if invalid_node_results:
                log.error("[%s] Backend provisioning yielded invalid node results:", pool.name)
                for node_res in invalid_node_results:
                    log.error("[%s]     %s", pool.name, node_res)

            log.debug("[%s] Setting hostname and ipaddr fields of nodes.", pool.name)
            for node, node_result in zip(nodes, valid_node_results):
                node.hostname = node_result["hostname"]
                node.ipaddr = node_result["ipaddr"]

        log.debug("[%s] Storing information about provisioned hosts.", pool.name)
        for node, node_result in zip(nodes, valid_node_results):
            if not reuse_nodes:
                node.hostname = node_result["hostname"]
                node.ipaddr = node_result["ipaddr"]
            node.data["provision"] = node_result
            node.state = NodeState.ready

        num_valid_nodes = len(valid_node_results)
        leftover_nodes = nodes[num_valid_nodes:]
        if leftover_nodes:
            if reuse_nodes:
                log.warning(
                    "[%s] Returning %d left-over reusable nodes", pool.name, len(leftover_nodes)
                )
                for node in leftover_nodes:
                    node.state = NodeState.unused
                    node.pool = None
                    node.data.pop("provision", None)
            else:
                log.warning(
                    "[%s] Cleaning up %d left-over preallocated nodes",
                    pool.name,
                    len(leftover_nodes),
                )
                for node in leftover_nodes:
                    db_sync_session.delete(node)

    log.info("[%s] Filling up nodes finished", pool.name)


@celery.task
def fill_pools(*, pool_names: Optional[List[str]] = None):
    """Ensure that pools are filled to their configured levels.

    If no pool names are supplied, run for all configured pools.
    """
    log.debug("fill_pools(pool_names=%r) begin", pool_names)

    pools_to_process = list(ConcreteNodePool.iter_pools())

    if not pool_names:
        pool_names = [pool.name for pool in pools_to_process]
    else:
        unknown_pool_names = set(pool_names).difference(pool.name for pool in pools_to_process)
        if unknown_pool_names:
            log.warn("fill_pools: unknown pool names, ignoring: %s", ", ".join(unknown_pool_names))
        pools_to_process = [pool for pool in pools_to_process if pool.name in pool_names]

    for pool in pools_to_process:
        fill_single_pool.delay(pool.name).forget()

    log.debug("fill_pools(%s) end", ", ".join(pool_names))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['duffy',
 'duffy.api_models',
 'duffy.app',
 'duffy.app.controllers',
 'duffy.configuration',
 'duffy.database',
 'duffy.database.migrations',
 'duffy.database.model',
 'duffy.legacy',
 'duffy.tasks',
 'duffy.tasks.mechanisms']

package_data = \
{'': ['*'], 'duffy.database.migrations': ['versions/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'SQLAlchemy>=1.4.25,<2.0.0',
 'alembic>=1.7.5,<2.0.0',
 'bcrypt>=3.2.0,<4.0.0',
 'celery[redis]>=5.2.1,<6.0.0',
 'click>=8.0.3,<9.0.0',
 'fastapi>=0.70.0,<0.71.0',
 'uvicorn>=0.15.0,<0.16.0']

extras_require = \
{'interactive': ['ipython>=7.29.0,<8.0.0'],
 'legacy': ['httpx>=0.18.2,<0.19.0'],
 'postgresql': ['psycopg2>=2.9.2,<3.0.0'],
 'tasks': ['ansible>=5.2.0,<6.0.0',
           'ansible-core>=2.12.1,<3.0.0',
           'ansible-runner>=2.1.1,<3.0.0',
           'Jinja2>=3.0.3,<4.0.0',
           'jmespath>=0.10.0,<0.11.0']}

entry_points = \
{'console_scripts': ['duffy = duffy.cli:cli']}

setup_kwargs = {
    'name': 'duffy',
    'version': '3.0.0a0',
    'description': 'CentOS CI provisioner',
    'long_description': "## Info\nCommunity Platform Engineering team (of Red Hat) is working on revamping this project and thus, have cleaned this repository by\n* marking other branches stale\n* Clean branch created for development\n\nto see the current deployed version of Duffy in CentOS CI Infra, check stale/master branch.\n\n\n## Duffy\nDuffy is the middle layer running ci.centos.org that manages the provisioning, maintenance and teardown / rebuild of the Nodes (physical hardware for now, VMs coming soon) that are used to run the tests in the CI Cluster.\n\n## Development\n\n### Installation\nTo install Duffy:\n1. Clone the repository and navigate into the project directory.\n   ```\n   git clone https://github.com/CentOS/duffy.git\n   cd duffy\n   ```\n2. Set up and activate a virtual environment.\n   * Using native virtual environment\n     ```\n     python3 -m venv duffyenv\n     source duffyenv/bin/activate\n     ```\n   Or\n   * Using virtualenv wrapper\n     ```\n     virtualenv duffyenv\n     source duffyenv/bin/activate\n     ```\n   Or\n   * Using Poetry virtual environment shell\n     ```\n     poetry shell\n     ```\n3. Install using Poetry\n   ```\n   poetry install\n   ```\n\n### Running Duffy server\n\n#### Viewing CLI usage\n\n```\nduffy --help\n```\n\n```\nUsage: duffy [OPTIONS]\n\n  Duffy is the middle layer running ci.centos.org that manages the\n  provisioning, maintenance and teardown / rebuild of the Nodes (physical\n  hardware for now, VMs coming soon) that are used to run the tests in the CI\n  Cluster.\n\nOptions:\n  -p, --portnumb INTEGER          Set the port value [0-65536]\n  -6, --ipv6                      Start the server on an IPv6 address\n  -4, --ipv4                      Start the server on an IPv4 address\n  -l, --loglevel [critical|error|warning|info|debug|trace]\n                                  Set the log level\n  --version                       Show the version and exit.\n  --help                          Show this message and exit.\n```\n\n#### Starting the server at port 8080 using IP version 4 and setting the log level to `trace`\n\n```\nduffy -p 8000 -4 -l trace\n```\n\n```\n * Starting Duffy...\n * Port number : 8000\n * IP version  : 4\n * Log level   : trace\nINFO:     Started server process [104283]\nINFO:     Waiting for application startup.\nTRACE:    ASGI [1] Started scope={'type': 'lifespan', 'asgi': {'version': '3.0', 'spec_version': '2.0'}}\nTRACE:    ASGI [1] Receive {'type': 'lifespan.startup'}\nTRACE:    ASGI [1] Send {'type': 'lifespan.startup.complete'}\nINFO:     Application startup complete.\nINFO:     Uvicorn running on http://0.0.0.0:8000 (Press CTRL+C to quit)\n```\n\nExit out of the server using `Ctrl` + `C`\n",
    'author': 'Nils Philippsen',
    'author_email': 'nils@redhat.com',
    'maintainer': 'Nils Philippsen',
    'maintainer_email': 'nils@redhat.com',
    'url': 'https://github.com/CentOS/duffy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import Enum
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union

from pydantic import AnyUrl, BaseModel, Field, conint, stricturl

# enums


class LogLevel(str, Enum):
    trace = "trace"
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


class MechanismType(str, Enum):
    ansible = "ansible"


# Pydantic models


class CeleryModel(BaseModel):
    broker_url: AnyUrl
    result_backend: AnyUrl


class PeriodicTaskModel(BaseModel):
    interval: conint(gt=0)


class TasksModel(BaseModel):
    celery: CeleryModel
    periodic: Optional[Dict[str, PeriodicTaskModel]]


class SQLAlchemyModel(BaseModel):
    sync_url: stricturl(tld_required=False, host_required=False)
    async_url: stricturl(tld_required=False, host_required=False)


class DatabaseModel(BaseModel):
    sqlalchemy: SQLAlchemyModel


class AnsibleMechanismPlaybookModel(BaseModel):
    extra_vars: Optional[Dict[str, Any]] = Field(alias="extra-vars")
    playbook: Path


class AnsibleMechanismModel(BaseModel):
    topdir: Optional[Path]
    extra_vars: Optional[Dict[str, Any]] = Field(alias="extra-vars")
    provision: Optional[AnsibleMechanismPlaybookModel]
    deprovision: Optional[AnsibleMechanismPlaybookModel]


class MechanismModel(BaseModel):
    type_: Optional[MechanismType] = Field(alias="type")
    ansible: Optional[AnsibleMechanismModel]


class NodePoolsModel(BaseModel):
    extends: Optional[str]
    mechanism: Optional[MechanismModel]
    fill_level: Optional[conint(gt=0)] = Field(alias="fill-level")
    reuse_nodes: Optional[Union[Dict[str, Union[int, str]], Literal[False]]] = Field(
        alias="reuse-nodes"
    )

    class Config:
        extra = "allow"


class NodePoolsRootModel(BaseModel):
    abstract: Optional[Dict[str, NodePoolsModel]]
    concrete: Dict[str, NodePoolsModel]


class LoggingModel(BaseModel):
    version: Literal[1]

    class Config:
        extra = "allow"


class AppModel(BaseModel):
    loglevel: Optional[LogLevel]
    host: Optional[str]
    port: Optional[conint(gt=0, lt=65536)]
    logging: Optional[LoggingModel]


class LegacyModel(BaseModel):
    host: Optional[str]
    port: Optional[conint(gt=0, lt=65536)]
    dest: Optional[str]
    loglevel: Optional[LogLevel]
    logging: Optional[LoggingModel]
    usermap: Dict[str, str]


class ConfigModel(BaseModel):
    app: Optional[AppModel]
    tasks: Optional[TasksModel]
    database: Optional[DatabaseModel]
    metaclient: Optional[LegacyModel]
    nodepools: Optional[NodePoolsRootModel]



from pathlib import Path
from typing import List, Union

import yaml

from ..util import merge_dicts
from .validation import ConfigModel

config = {}


def read_configuration(*config_files: List[Union[Path, str]], clear: bool = True):
    new_config = {}
    for config_file in config_files:
        if not isinstance(config_file, Path):
            config_file = Path(config_file)
        with config_file.open("r") as fp:
            for config_doc in yaml.safe_load_all(fp):
                # validate configuration file
                ConfigModel(**config_doc)

                new_config = merge_dicts(new_config, config_doc)

    if clear:
        config.clear()

    config.update(new_config)



"""This is the tenant controller."""

from uuid import uuid4

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.status import (
    HTTP_201_CREATED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
)

from ...api_models import (
    TenantCreateModel,
    TenantCreateResult,
    TenantCreateResultModel,
    TenantModel,
    TenantResult,
    TenantResultCollection,
)
from ...database.model import Tenant
from ..auth import req_tenant
from ..database import req_db_async_session

router = APIRouter(prefix="/tenants")


# http get http://localhost:8080/api/v1/tenants
@router.get("", response_model=TenantResultCollection, tags=["tenants"])
async def get_all_tenants(
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Return all tenants."""
    query = select(Tenant)
    if not tenant.is_admin:
        query = query.filter_by(id=tenant.id)
    results = await db_async_session.execute(query)

    return {"action": "get", "tenants": results.scalars().all()}


# http get http://localhost:8080/api/v1/tenants/2
@router.get("/{id}", response_model=TenantResult, tags=["tenants"])
async def get_tenant(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Return the tenant with the specified **ID**."""
    retrieved_tenant = (
        await db_async_session.execute(select(Tenant).filter_by(id=id))
    ).scalar_one_or_none()

    if not retrieved_tenant:
        raise HTTPException(HTTP_404_NOT_FOUND)

    if retrieved_tenant != tenant and not tenant.is_admin:
        raise HTTPException(HTTP_403_FORBIDDEN)

    return {"action": "get", "tenant": retrieved_tenant}


# http --json post http://localhost:8080/api/v1/tenants name="Unique name"
@router.post("", status_code=HTTP_201_CREATED, response_model=TenantCreateResult, tags=["tenants"])
async def create_tenant(
    data: TenantCreateModel,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Create a tenant with the specified **name**."""
    if not tenant.is_admin:
        raise HTTPException(HTTP_403_FORBIDDEN)

    api_key = uuid4()

    created_tenant = Tenant(
        name=data.name,
        is_admin=data.is_admin,
        api_key=api_key,
        ssh_key=data.ssh_key.get_secret_value(),
    )
    db_async_session.add(created_tenant)
    try:
        await db_async_session.flush()
    except IntegrityError as exc:
        raise HTTPException(HTTP_409_CONFLICT, str(exc))

    api_tenant = TenantCreateResultModel(
        api_key=api_key, **TenantModel.from_orm(created_tenant).dict()
    )

    await db_async_session.commit()

    return {"action": "post", "tenant": api_tenant}


# http delete http://localhost:8080/api/v1/tenant/2
@router.delete("/{id}", response_model=TenantResult, tags=["tenants"])
async def delete_tenant(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Delete the tenant with the specified **ID**."""
    if not tenant.is_admin:
        raise HTTPException(HTTP_403_FORBIDDEN)

    tenant = (await db_async_session.execute(select(Tenant).filter_by(id=id))).scalar_one_or_none()

    if not tenant:
        raise HTTPException(HTTP_404_NOT_FOUND)

    await db_async_session.delete(tenant)
    await db_async_session.commit()

    return {"action": "delete", "tenant": tenant}



import uuid

import bcrypt
from sqlalchemy import Boolean, Column, Integer, Text, UnicodeText, text
from sqlalchemy.ext.hybrid import hybrid_property

from .. import Base
from ..util import CreatableMixin, RetirableMixin


class Tenant(Base, CreatableMixin, RetirableMixin):
    __tablename__ = "tenants"
    __mapper_args__ = {"eager_defaults": True}
    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(UnicodeText, unique=True, nullable=False)
    is_admin = Column(Boolean, nullable=False, default=False, server_default=text("FALSE"))
    ssh_key = Column(UnicodeText, nullable=False)
    _api_key = Column("api_key", Text, nullable=False)

    @hybrid_property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, key: uuid.UUID):
        salt = bcrypt.gensalt()
        hashed_key = bcrypt.hashpw(str(key).encode("ascii"), salt)
        self._api_key = hashed_key.decode("ascii")

    def validate_api_key(self, key: uuid.UUID):
        return bcrypt.checkpw(str(key).encode("ascii"), self._api_key.encode("ascii"))
